.grid {
  display: grid;
  grid-template-columns: auto auto;
  grid-gap: 1em;
}

.grid .highlight {
  display: flex;
  flex-direction: column;
}

.grid pre {
  flex: 1;
  display: flex;
  flex-direction: column;
}

.grid code {
  flex: 1;
}
