"""update tables

Revision ID: 0.3.4
Revises: 0.3.3
Create Date: 2023-03-14 12:09:29.232008

"""

# revision identifiers, used by Alembic.
revision = "0.3.4"
down_revision = "0.3.3"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###
