"""update tables

Revision ID: 0.6.5
Revises: 0.6.4
Create Date: 2024-07-09 16:29:09.389585

"""

# revision identifiers, used by Alembic.
revision = "0.6.5"
down_revision = "0.6.4"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###
