"""update tables

Revision ID: 0.6.2
Revises: 0.6.1
Create Date: 2023-08-18 17:11:15.605131

"""

# revision identifiers, used by Alembic.
revision = "0.6.2"
down_revision = "0.6.1"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###
