"""update tables

Revision ID: 0.5.3
Revises: 0.5.2
Create Date: 2023-06-13 10:27:02.691266

"""


# revision identifiers, used by Alembic.
revision = "0.5.3"
down_revision = "0.5.2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###
