"""file_add_composite_index_dataset_id_status

Revision ID: d14f179e553c
Revises: e7edab5d4e4b
Create Date: 2025-06-18 16:05:35.721085

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd14f179e553c'
down_revision = 'e7edab5d4e4b'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('file', schema=None) as batch_op:
        batch_op.create_index('ix_file_dataset_status', ['dataset_id', 'status'], unique=False)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('file', schema=None) as batch_op:
        batch_op.drop_index('ix_file_dataset_status')

    # ### end Alembic commands ###
